#!/usr/bin/env python
# -*- coding: utf-8 -*-

import datetime
from classes import dfMenu_MD
from classes import dfMenu_SFC
from classes import dfMenu_SMSGG

# トップメニュー
menuStr_Top = \
            '\n'\
            '==========================================================\n'\
            '                     <<トップメニュー>>                   \n'\
            '==========================================================\n'\
            '接続中カセット種別: %s                                    \n'\
            '----------------------------------------------------------\n'\
            '1: メニュー：メガドライブ                                 \n'\
            '2: メニュー：スーパーファミコン                           \n'\
            '3: メニュー：マークIII／ゲームギア                        \n'\
            'q: 終了                                                   \n'\
            '----------------------------------------------------------\n'\
            '選択してください: '


#---------------------------------------------------------------------------
# 何のゲームのカセットが刺さっているかを返す
#   戻り値:2値のタプル
#       stateStr:  状態文字列
#       connected: 何かカセットが刺さっているか(0:なし/1:カセットあり)
#---------------------------------------------------------------------------
def getCartStateStr(s):
    stateStr = ''

    # 状態要求コード(?)送信
    s.write(bytes([
        0x06, 0x00, 0x00, 0x00,
        0x00, 0x00, 0x00, 0x00,
        0x04, 0x00, 0x00, 0x00,]))

    # 応答受信
    readStr = s.read(1024)

    # 応答解析
    readBytes = bytes(readStr)
    
    connected = 0
    if len(readBytes) == 0:
        stateStr = 'ERR: カートリッジアダプタが応答しません'
        connected = 0
    elif len(readBytes) < 12:
        stateStr = 'ERR: カートリッジアダプタの応答が不正です。'
        connected = 0
    else:
        #print("%s 0x%02x" % (datetime.datetime.now(), readBytes[8]), end='')
        
        if   readBytes[8] == 0xff:
            stateStr = '未接続'
            connected = 0

        elif readBytes[8] == 0x01:
            stateStr = 'スーパーファミコン'
            connected = 1

        elif readBytes[8] == 0x02:
            stateStr = 'メガドライブ'
            connected = 1

        elif readBytes[8] == 0x03:
            stateStr = 'ファミコン'
            connected = 1

        elif readBytes[8] == 0x05:
            stateStr = 'ゲームボーイ'
            connected = 1

        elif readBytes[8] == 0x07:
            stateStr = 'マークIII／ゲームギア'
            connected = 1

        elif readBytes[8] == 0x08:
            stateStr = 'PCエンジンHuカード'
            connected = 1

    return (stateStr, connected)

#---------------------------------------------------------------------------
#
# メニュー：トップ
#
#---------------------------------------------------------------------------
def menuProc_Top(s, stateStr, connected, g_menu):
    ## 書き換え予定のグローバル変数宣言
    #global g_menu
    
    # メニュー表示
    print(menuStr_Top % stateStr, end='')
    
    # 入力待ち＆判定
    inStr = input()
    if len(inStr) == 0:
        return g_menu

    elif inStr == 'q':      # 終了
        print('終了します。')
        g_menu.pop()
        return g_menu

    elif inStr == '1':      # メニュー：メガドライブ
        g_menu.append(1)
        return g_menu

    elif inStr == '2':      # メニュー：スーパーファミコン
        g_menu.append(2)
        return g_menu

    elif inStr == '3':      # メニュー：マークIII／ゲームギア
        g_menu.append(3)
        return g_menu

    else:                   # 不明コマンド
        print('不明なコマンドです。')
        return g_menu

#---------------------------------------------------------------------------
# メニュー ループ処理
#---------------------------------------------------------------------------
def menuLoop(s):
    g_menu=[0]

    # ループ
    while True:
        # RFCA状態取得
        (stateStr, connected) = getCartStateStr(s)
        
        # print('g_menu', g_menu)
        if g_menu == [0]:
            # メニュー：トップ
            g_menu = menuProc_Top(s, stateStr, connected, g_menu)

        elif g_menu == [0, 1]:
            # メニュー：メガドライブ
            g_menu = dfMenu_MD.menuProc(s, stateStr, connected, g_menu)

        elif g_menu == [0, 2]:
            # メニュー：スーパーファミコン
            g_menu = dfMenu_SFC.menuProc(s, stateStr, connected, g_menu)

        elif g_menu == [0, 3]:
            # メニュー：マークIII／ゲームギア
            g_menu = dfMenu_SMSGG.menuProc(s, stateStr, connected, g_menu)

        elif g_menu == []:
            # 終了
            break
    
    # End of while

