#!/usr/bin/env python
# -*- coding: utf-8 -*-

from classes import df_Common
import struct

# デバッグ用の冗長なログ出力 0:OFF,1:ON
verbose = 0

# メニュー：メガドライブ
menuStr_MD = \
            '\n'\
            '==========================================================\n'\
            '               <<メニュー：メガドライブ>>                 \n'\
            '==========================================================\n'\
            '接続中カセット種別: %s                                    \n'\
            '----------------------------------------------------------\n'\
            '1: 先頭512バイトをダンプ表示                              \n'\
            '2: 指定アドレスから512バイトダンプ表示                    \n'\
            '3: 指定アドレスから指定サイズダンプ表示                   \n'\
            '4: 16MbitメガドラROM読み出し＆ROMファイル作成             \n'\
            '5: 指定サイズROM読み出し＆ROMファイル作成                 \n'\
            'b: 戻る                                                   \n'\
            'q: 終了                                                   \n'\
            '----------------------------------------------------------\n'\
            '選択してください: '



#---------------------------------------------------------------------------
#
# メニュー：メガドライブ
#
#---------------------------------------------------------------------------
def menuProc(s, stateStr, connected, g_menu):
    ## 書き換え予定のグローバル変数宣言
    #global g_menu
    
    # メニュー表示
    print(menuStr_MD % stateStr, end='')
    
    # 入力待ち＆判定
    inStr = input()
    if len(inStr) == 0:
        return g_menu

    elif inStr == 'b':      # 戻る
        g_menu.pop()
        return g_menu

    elif inStr == 'q':      # 終了
        g_menu = []
        return g_menu

    elif inStr == '1':      # 先頭512バイトダンプ
        if connected != 1:
            print('カセットが未接続のためダンプできません。')
        else:
            print('先頭ダンプ')
            (retCode, retAry) = df_Common.readSize5(s, 0, 512)
            print('retCode: %d' % retCode)
            df_Common.dump(retAry)
            print('')
        return g_menu

    elif inStr == '2':      # 指定アドレス512バイトダンプ
        if connected != 1:
            print('カセットが未接続のためダンプできません。')
        else:
            print('指定アドレスダンプ')

            # アドレスを入力させる
            stAddr = df_Common.inputHexValue('開始アドレス(16進)を入力してください。: ')

            (retCode, retAry) = df_Common.readSize5(s, stAddr, 512)
            df_Common.dump(retAry, stAddr)
            print('')
        return g_menu

    elif inStr == '3':      # 指定アドレス指定サイズダンプ
        if connected != 1:
            print('カセットが未接続のためダンプできません。')
        else:
            df_Common.putLog('指定アドレス指定サイズダンプ')

            # アドレスを入力させる
            stAddr = df_Common.inputHexValue('開始アドレス(16進)を入力してください。: ')
            
            # サイズを入力させる
            stSize = df_Common.inputHexValue('サイズ(16進)を入力してください。: ')

            (retCode, retAry) = df_Common.readSize5(s, stAddr, stSize)
            df_Common.dump(retAry, stAddr)
            print('')
        return g_menu

    elif inStr == '4':      # 16Mbitメガドラカセットファイル化
        if connected != 1:
            print('カセットが未接続のためダンプできません。')
        else:
            df_Common.putLog('16Mbitメガドラカセットファイル化')

            # ファイル名入力
            fileNameStr = df_Common.inputFilename('作成ファイル名を入力してください。: ')
            
            fileAry = []
            addr = 0
            dumpSize = 1024
            #dumpSize = 512
            while addr < 1024*1024*2:
                # ファイル出力用配列にバンバン追加
                print('\rアドレス {:08x} から読み出し中...'.format(addr), end='')
                (retCode, retAry) = df_Common.readSize5(s, addr, dumpSize)
                if retCode == 0:
                    fileAry.extend(retAry)
                    addr = addr + dumpSize
                else:
                    print('リトライ実施(retCode: %d)' % retCode)
                    #s.rtscts = 1
                    continue
            print('完了。')
            
            # 配列をファイル出力
            with open(fileNameStr, "wb") as fOut:
                for x in fileAry:
                    fOut.write(struct.pack("B", x))
            print('ファイル: %sを出力しました。' % fileNameStr)
            print('')
        return g_menu

    elif inStr == '5':      # 指定サイズメガドラカセットファイル化
        if connected != 1:
            print('カセットが未接続のためダンプできません。')
        else:
            df_Common.putLog('指定サイズメガドラカセットファイル化')

            # ファイル名入力
            fileNameStr = df_Common.inputFilename('作成ファイル名を入力してください。: ')
            
            # サイズ入力
            endAddr = df_Common.inputHexValue('サイズ(16進)を入力してください。: ')

            fileAry = []
            addr = 0
            dumpSize = 1024
            #dumpSize = 512
            while addr < endAddr:
                # ファイル出力用配列にバンバン追加
                print('\rアドレス {:08x} から読み出し中...'.format(addr), end='')
                (retCode, retAry) = df_Common.readSize5(s, addr, dumpSize)
                if retCode == 0:
                    fileAry.extend(retAry)
                    addr = addr + dumpSize
                else:
                    print('リトライ実施(retCode: %d)' % retCode)
                    #s.rtscts = 1
                    continue
            print('完了。')
            
            # 配列をファイル出力
            with open(fileNameStr, "wb") as fOut:
                for x in fileAry:
                    fOut.write(struct.pack("B", x))
            print('ファイル: %sを出力しました。' % fileNameStr)
            print('')
        return g_menu

    else:                   # 不明コマンド
        print('不明なコマンドです。')
        return g_menu
