#!/usr/bin/env python
# -*- coding: utf-8 -*-

import struct
import time     # sleep
from classes import df_Common
from classes import df_SFC
from classes import df_SFM

# メニュープログラム内でややこしい処理が増えてきたら、
# 個別ゲーム用クラスに移す


# デバッグ用の冗長なログ出力 0:OFF,1:ON
verbose = 0

# メニュー：スーパーファミコン
menuStr_SFC = \
            '\n'\
            '==========================================================\n'\
            '             <<メニュー：スーパーファミコン>>             \n'\
            '==========================================================\n'\
            '接続中カセット種別: %s                                    \n'\
            '----------------------------------------------------------\n'\
            ' 1: 先頭512バイトをダンプ表示                              \n'\
            ' 2: 指定アドレスから512バイトダンプ表示                    \n'\
            ' 3: カートリッジヘッダー表示                               \n'\
            ' 4: HiROM各バンク先頭＆末尾ダンプ                          \n'\
            ' 5: ROM読み出し＆ROMファイル作成                           \n'\
            ' 6: [SFM]SFメモリカセットWakeup(HIROM:ALL)                 \n'\
            ' 7: [SFM]ディレクトリ一覧表示                              \n'\
            ' 8: [SFM]Wakeup＆全体吸出し＆ROMファイル作成               \n'\
            ' 9: [SFM]Wakeup＆ゲーム選択吸出し＆ROMファイル作成         \n'\
            '10: [SFM]Wakeup＆マップ情報表示                            \n'\
            '11: [SFM]Wakeup＆マップ情報ファイル保存                    \n'\
            '12: [SFM]SFメモリカセット用イメージ書き込み                \n'\
            '13: [SFM]マップ情報書き込み                                \n'\
            '14: [SFM]全セクターのプロテクト状態表示                    \n'\
            '----------------------------------------------------------\n'\
            '                    [※！警告！※]                         \n'\
            '   以下の処理は、SFメモリカセットが復旧できない状態になる  \n'\
            '   危険性があります。                                      \n'\
            '----------------------------------------------------------\n'\
            '15: [SFM]マップ情報イレーズコマンド送信                    \n'\
            '16: [SFM]c0/e0リセットコマンド送信                         \n'\
            '17: [SFM]Flash c0セクタープロテクト有効化                  \n'\
            '18: [SFM]Flash c0末尾セクタープロテクト無効化              \n'\
            '19: [SFM]Flash c0プロテクト有効化                          \n'\
            '20: [SFM]Flash c0プロテクト無効化                          \n'\
            'b: 戻る                                                   \n'\
            'q: 終了                                                   \n'\
            '----------------------------------------------------------\n'\
            '選択してください: '




#---------------------------------------------------------------------------
#
# メニュー：スーパーファミコン
#
#---------------------------------------------------------------------------
def menuProc(s, stateStr, connected, g_menu):
    ## 書き換え予定のグローバル変数宣言
    #global g_menu
    
    # メニュー表示
    print(menuStr_SFC % stateStr, end='')
    
    # 入力待ち＆判定
    inStr = input()
    if len(inStr) == 0:
        return g_menu

    elif inStr == 'b':      # 戻る
        g_menu.pop()
        return g_menu

    elif inStr == 'q':      # 終了
        g_menu = []
        return g_menu

    elif inStr == '1':      # 先頭512バイトダンプ
        MenuItem_SFC_DumpTop512(s, connected)
        return g_menu

    elif inStr == '2':      # 指定アドレス512バイトダンプ
        MenuItem_SFC_DumpFromSomewhare512(s, connected)
        return g_menu

    elif inStr == '3':      # カートリッジヘッダー表示
        MenuItem_SFC_DisplayCartHeader(s, connected)
        return g_menu

    elif inStr == '4':
        # 各バンク先頭＆末尾ダンプ
        # HiROMと見なして、カセット全体を俯瞰するために作った・・・？
        MenuItem_SFC_DisplayAllBanksTopAndBottom(s, connected)
        return g_menu
        

    elif inStr == '5':      
        # ROM読み出し＆ROMファイル作成
        MenuItem_SFC_ReadROMAndCreateFile(s, connected)
        return g_menu
        # メニュー5

    elif inStr == '6':      # SFメモリカセットWakeup
        MenuItem_SFC_SFM_Wakeup(s, connected)
        return g_menu

    elif inStr == '7':      # [SFM]ディレクトリ一覧表示
        MenuItem_SFC_SFM_DisplayDir(s, connected)
        return g_menu
        # メニュー7

    elif inStr == '8':      
        # [SFM]Wakeup＆全体吸出し＆ROMファイル作成
        MenuItem_SFC_SFM_ReadROMAndCreateFile_ALL(s, connected)
        return g_menu
        # ここまでメニュー8

    elif inStr == '9':      
        # [SFM]Wakeup＆ゲーム選択吸出し＆ROMファイル作成
        MenuItem_SFC_SFM_ReadROMAndCreateFile_Selected(s, connected)
        return g_menu
        # ここまでメニュー9

    elif inStr == '10':
        # [SFM]Wakeup＆マップ情報表示
        MenuItem_SFC_SFM_DisplayMap(s, connected)
        return g_menu
        # ここまでメニュー10

    elif inStr == '11':
        # [SFM]Wakeup＆マップ情報ファイル保存
        MenuItem_SFC_SFM_ReadMapInfoAndSaveFile(s, connected)
        return g_menu
        # ここまでメニュー11

    elif inStr == '12':     
        # [SFM]SFメモリカセット用ROMイメージ書き込み
        MenuItem_SFC_SFM_WriteROMImage(s, connected)
        return g_menu
        # ここまでメニュー12

    elif inStr == '13':
        # [SFM]マップ情報書き込み
        MenuItem_SFC_SFM_WriteMAPInfo(s, connected)
        
        return g_menu
        # ここまでメニュー13

    elif inStr == '14':      
        # 全セクターのプロテクト状態表示
        MenuItem_SFC_SFM_DispSectorProtect(s, connected)
        return g_menu
        # メニュー14

    elif inStr == '15':      
        # [SFM] マップ情報イレーズコマンド送信＆コマンド終了まで待機
        MenuItem_SFC_SFM_EraseMapInfo(s, connected)
        return g_menu
        # メニュー15

    elif inStr == '16':      
        # [SFM]c0/e0リセットコマンド送信
        MenuItem_SFC_SFM_Reset(s, connected)
        return g_menu
        # メニュー16

    elif inStr == '17':      
        # [SFM] c0&c1セクタープロテクト有効化
        MenuItem_SFC_SFM_SectorProtectC0(s, connected)
        return g_menu
        # メニュー17

    elif inStr == '18':      
        # [SFM]Flash c0末尾セクタープロテクト無効化
        MenuItem_SFC_SFM_SectorUnprotectC0End(s, connected)
        return g_menu
        # メニュー18

    elif inStr == '19':      
        # [SFM]Flash c0プロテクト有効化
        MenuItem_SFC_SFM_SetFlashProtect(s, connected)
        return g_menu
        # メニュー19

    elif inStr == '20':
        # [SFM]Flash c0プロテクト無効化
        MenuItem_SFC_SFM_UnsetFlashProtect(s, connected)
        return g_menu
        # メニュー20


    else:                   # 不明コマンド
        print('不明なコマンドです。')
        return g_menu





#------------------------------------------------
#
# メニューアイテム
# 先頭512バイトダンプ
#
#------------------------------------------------
def MenuItem_SFC_DumpTop512(s, connected):
    if connected != 1:
        print('カセットが未接続のためダンプできません。')
    else:
        print('先頭ダンプ')
        (retCode, retAry) = df_Common.readSizeByType(s, 0x07, 0, 512)
        print('retCode: %d' % retCode)
        df_Common.dump(retAry)
        print('')

#------------------------------------------------
#
# メニューアイテム
# 指定アドレス512バイトダンプ
#
#------------------------------------------------
def MenuItem_SFC_DumpFromSomewhare512(s, connected):
    if connected != 1:
        print('カセットが未接続のためダンプできません。')
    else:
        print('指定アドレスダンプ')

        # アドレスを入力させる
        stAddr = df_Common.inputHexValue('開始アドレス(16進)を入力してください。: ')

        (retCode, retAry) = df_Common.readSizeByType(s, 0x07, stAddr, 512)
        df_Common.dump(retAry, stAddr)
        print('')

#------------------------------------------------
#
# メニューアイテム
# SFメモリカセットWakeup
#
#------------------------------------------------
def MenuItem_SFC_SFM_Wakeup(s, connected):
    if connected != 1:
        print('カセットが未接続のため操作できません。')
    else:
        print('SFメモリカセットWakeup')

    if df_SFM.wakeup_SFM(s) == 0:
        print('SFメモリカセットWakeup処理に成功しました。')
    else:
        print('Err: SFメモリカセットWakeup処理に失敗しました。')
    
#------------------------------------------------
# メニューアイテム
# [SFM]Wakeup＆マップ情報表示
#
#------------------------------------------------
def MenuItem_SFC_SFM_DisplayMap(s, connected):
    # [SFM]Wakeup＆マップ情報表示
    if connected != 1:
        print('カセットが未接続のため処理できません。')
        return

    print('[SFM]Wakeup＆マップ情報表示')
    
    # SFメモリwakeup処理
    df_SFM.wakeup_SFM(s)
    
    print('0.3秒待機します。')
    time.sleep(0.3)

    # Wakeup状態チェック
    (retCode, retAry) = df_Common.readSizeByType(s, 0x07, 0x2400, 16)
    if retCode != 0:
        print('Err: Wakeup状態読み取り処理失敗。')
        return
    else:
        if retAry[0] == 0x2a:
            print('SFメモリカセットのWakeup処理に成功しました。')
        else:
            print('Err: SFメモリカセットのWakeup処理に失敗しました。')
            return
            
    # ここまで来たら、Wakeup(HIMEM:ALL)に成功している

    # Maker ID/Device IDの確認
    (mID_c0, dID_c0) = df_SFM.idFlash_SFM(s, 0xc0)
    if (mID_c0 != 0xc2) or (dID_c0 != 0xf3):
        print('Err: Flash %02x Maker ID/Device IDが不正です。(%02x%02x)' % (0xc0, mID_c0, dID_c0))
        return
    print('Flash %02x Maker ID/Device ID: %02x/%02x...OK' % (0xc0, mID_c0, dID_c0))

    (mID_e0, dID_e0) = df_SFM.idFlash_SFM(s, 0xe0)
    if (mID_e0 != 0xc2) or (dID_e0 != 0xf3):
        print('Err: Flash %02x Maker ID/Device IDが不正です。(%02x%02x)' % (0xe0, mID_e0, dID_e0))
        return
    print('Flash %02x Maker ID/Device ID: %02x/%02x...OK' % (0xe0, mID_e0, dID_e0))

    # リセット
    print('Flashをリセットし、0.1秒待機します。')
    df_SFM.resetFlash_SFM(s, 0xc0)
    df_SFM.resetFlash_SFM(s, 0xe0)
    time.sleep(0.1)
    
    # マップ情報
    print('マップ情報（ダンプ）')
    df_SFM.printMapping_SFM(s)
    
    print('Enterキーを押してください。')
    input()
    
    return

#------------------------------------------------
#
# メニューアイテム
# [SFM]ディレクトリ一覧表示
#
#------------------------------------------------
def MenuItem_SFC_SFM_DisplayDir(s, connected):
    if connected != 1:
        print('カセットが未接続のため操作できません。')
    else:
        print('[SFM]ディレクトリ一覧表示')
        
        # ここから送信処理

        # 1-1. 2Fコマンド送信
        df_Common.putLog('2Fコマンド送信')
        s.write(bytes([
            0x2f, 0x00, 0x00, 0x00,             # 
            0x08, 0x00, 0x00, 0x00,             # 不明 8
            0x00, 0x00, 0x00, 0x00,             # 不明 0
            0x00, 0x00, 0x00, 0x00,             # 不明 0
            0x01, 0x00, 0x00, 0x00              # 不明 1
            ]))

        # 1-2. 2Fコマンド応答待ち
        respWaitCount = 0
        while True:
            if s.in_waiting > 0:
                df_Common.putLog('受信データあり: %d' % s.in_waiting)
                break
            else:
                respWaitCount = respWaitCount + 1
                df_Common.putLog('0.01秒待機。')
                time.sleep(0.01)
                if(respWaitCount > 40):
                    print('ERR: 2Fコマンド応答タイムアウト')
                    return

        # 1-3. 2Fコマンド応答の受信と判定
        readStr = s.read(8)
        readBytes = bytes(readStr)
        df_Common.putLog('2Fコマンド応答受信。')
        df_Common.debugDump(readBytes)

        if readBytes == bytes([0,0,0,0,0,0,0,0]):
            df_Common.putLog('肯定応答を受信。')
        else:
            #df_Common.putLog('肯定応答でない応答を受信。')
            #print('ERR: 2Fコマンドに肯定応答でない応答')
            #return -2, retAry
            df_Common.putLog('肯定応答でない応答を受信。続行。')

        time.sleep(0.5)

        enableCount = 0
        while True:
            df_Common.putLog('SFメモリWakeupコマンド(HIROM:ALL)送信')
            df_SFM.send_SFM(s, 0x04)

            (retCode, retAry) = df_Common.readSizeByType(s, 0x07, 0x2400, 1)

            if (retCode == 0):
                df_Common.putLog('Wakeupコマンド完了後の0x2400の値: 0x%02x' % retAry[0])
                if retAry[0] == 0x2a:
                    df_Common.putLog('Wakeupコマンド正常終了')
                    break
                else:
                    time.sleep(0.1)
                    enableCount = enableCount + 1
                    if enableCount > 5:
                        df_Common.putLog('Err: リトライ上限到達。続行')
                        break
                    else:
                        df_Common.putLog('Err: コマンド完了後の0x2400の値!=0x2a。0.1秒後リトライ実施。')
                        time.sleep(0.1)
                        continue
            else:
                enableCount = enableCount + 1
                if enableCount > 5:
                    df_Common.putLog('Err: リトライ上限到達。続行')
                    break
                else:
                    df_Common.putLog('Err: コマンド完了後の0x2400の値の取得に失敗。0.1秒後リトライ実施。')
                    time.sleep(0.1)
                    continue

        # ここまで2F送信＆有効化処理
        
        # コマンド0x8nが効けば、MENUPROGRAMに切り替えて
        # 0x8c8000からディレクトリ一覧を表示すべきだった。
        # が、うまくコマンドが実行できない。
        # このため、HIROM:ALL状態でMENUPROGRAMを探し、
        # ディレクトリ情報を直に読んでしまう。
        #df_Common.putLog('コマンド0x80送信')
        #df_SFM.send_SFM(s, 0x80)
        
        # MENUプログラム存在チェック
        (retCode, retAry) = df_Common.readSizeByType(s, 0x07, 0xc07fc0, 12)
        if retCode != 0:
            print('Err: 読み取り処理失敗。')
            return
        # 'MENU PROGRAM'か？
        if bytes(retAry) == bytes([0x4D, 0x45, 0x4E, 0x55, 
                                   0x20, 0x50, 0x52, 0x4F, 
                                   0x47, 0x52, 0x41, 0x4D]):
            print('MENU PROGRAMを検知しました。')
            hasMenu = 1
        else:
            print('MENU PROGRAMが存在しないか、検知できませんでした。')
            return
        
        # ディレクトリ一覧表示
        print('#  F  B Gamecode     タイトル                                     Date       Time     LAW/NIN ')
        print('- -- -- ------------ -------------------------------------------- ---------- -------- --------')
        #      0 00 00 123456789012 ０１２３４５６７８９０１２３４５６７８９０１ yyyy/mm/dd hh:mm:ss LAWxxxxx
        i = 0
        addr = 0xc60000
        while i < 8:
            # ディレクトリ位置0の情報を持ってくる
            (retCode, retAry) = df_Common.readSizeByType(s, 0x07, addr, 512)
            
            if retCode != 0:
                print('Err: ディレクトリ情報参照失敗。')
                break
            else:
                ## test
                #df_Common.dump(retAry, addr)
                
                if retAry[0] >= 8:
                    # ディレクトリ情報はもう無い
                    break
                
                # 各種情報の計算
                dirIdx = retAry[0]
                numOfFlashBlock = (retAry[4] * 0x100 + retAry[3]) / 4
                numOfSRAMBlock  = (retAry[6] * 0x100 + retAry[5]) / 16
                gamecode = str(bytes(retAry[7:19]), encoding='utf-8', errors='replace')
                sjisName = str(bytes(retAry[0x13:0x3f]), encoding='sjis', errors='replace')
                dateStr  = str(bytes(retAry[0x1bf:0x1c9]), encoding='utf-8', errors='replace')
                timeStr  = str(bytes(retAry[0x1c9:0x1d1]), encoding='utf-8', errors='replace')
                lawnin   = str(bytes(retAry[0x1d1:0x1d9]), encoding='utf-8', errors='replace')
                
                print('%1d %2d %2d %12s %s %s %s %s' % 
                    (dirIdx, numOfFlashBlock, numOfSRAMBlock, gamecode, sjisName,
                     dateStr, timeStr, lawnin))
                
                # 次のディレクトリ情報位置へ
                i = i + 1
                addr = addr + 0x2000
        
        # 入力待ち
        print('Enterキーを押してください。')
        inStr = input()

#------------------------------------------------
#
# メニューアイテム
# カートリッジヘッダー表示
#
#------------------------------------------------
def MenuItem_SFC_DisplayCartHeader(s, connected):
    # カートリッジヘッダー表示
    if connected != 1:
        print('カセットが未接続のため処理できません。')
    else:
        print('カートリッジヘッダー表示')

        # ROM種別(HiROM, LoROM, ExHiROM)判定
        # この時点ではHiROM, LoROM, ExHiROMの区別がついていない。
        # ヘッダ情報が入っているはずの領域をアクセスし、
        # チェックサム部から判定する
        (ret, titleStr, speedVal, mapModeVal, mapModeStr, romSizeVal, ramSizeVal, csum, csumComp) = df_SFC.get_SFCHeaderInfo(s)
        if ret == -1:
            print('Err: 処理中にエラー発生。')
            return g_menu
        else:
            print('Title:     %s' % titleStr)
            print('ROM Speed: %d (0=Slow, 1=Fast) (Slow 200ns, Fast 120ns)' % speedVal)
            print('Map mode:  %d' % mapModeVal)
            print('           %s' % mapModeStr)
            print('ROM size:  %d Kbytes' % romSizeVal)
            print('RAM size:  %d Kbytes' % ramSizeVal)
            print('Checksum complement:  0x%04x' % csumComp)
            print('Checksum           :  0x%04x' % csum)

        # 入力待ち
        print('Enterキーを押してください。')
        inStr = input()
        

#------------------------------------------------
#
# メニューアイテム
# 各バンク先頭＆末尾ダンプ
#
#------------------------------------------------
def MenuItem_SFC_DisplayAllBanksTopAndBottom(s, connected):
    if connected != 1:
        print('カセットが未接続のため処理できません。')
    else:
        print('各バンク先頭＆末尾ダンプ')
        
        bank = 0x40
        while bank < 0x7e:
            print('bank 0x%02x:' % bank)
            (retCode, retAry) = df_Common.readSizeByType(s, 0x07, 0      + bank * 0x10000, 16)
            df_Common.dump(retAry, bank * 0x10000)
            (retCode, retAry) = df_Common.readSizeByType(s, 0x07, 0xfff0 + bank * 0x10000, 16)
            df_Common.dump(retAry, bank * 0x10000 + 0xfff0)
            bank = bank + 1
            print('-----------------------------------')

        bank = 0xfe
        while bank < 0x100:
            print('bank 0x%02x:' % bank)
            (retCode, retAry) = df_Common.readSizeByType(s, 0x07, 0      + bank * 0x10000, 16)
            df_Common.dump(retAry, bank * 0x10000)
            (retCode, retAry) = df_Common.readSizeByType(s, 0x07, 0xfff0 + bank * 0x10000, 16)
            df_Common.dump(retAry, bank * 0x10000 + 0xfff0)
            bank = bank + 1
            print('-----------------------------------')
            
        # 入力待ち
        print('Enterキーを押してください。')
        inStr = input()
                


#------------------------------------------------
#
# メニューアイテム
# ROM読み出し＆ROMファイル作成
#
#------------------------------------------------
def MenuItem_SFC_ReadROMAndCreateFile(s, connected):
    if connected != 1:
        print('カセットが未接続のため処理できません。')
    else:
        print('ROM読み出し＆ROMファイル作成')

        print('ヘッダ情報を表示します。')

        # ROM種別(HiROM, LoROM, ExHiROM)判定
        # この時点ではHiROM, LoROM, ExHiROMの区別がついていない。
        # ヘッダ情報が入っているはずの領域をアクセスし、
        # チェックサム部から判定する
        (ret, titleStr, speedVal, mapModeVal, mapModeStr, romSizeVal, ramSizeVal, csum, csumComp) = df_SFC.get_SFCHeaderInfo(s)
        if ret == -1:
            print('Err: 処理中にエラー発生。')
            return g_menu
        else:
            print('Title:     %s' % titleStr)
            print('ROM Speed: %d (0=Slow, 1=Fast) (Slow 200ns, Fast 120ns)' % speedVal)
            print('Map mode:  %d' % mapModeVal)
            print('           %s' % mapModeStr)
            print('ROM size:  %d Kbytes' % romSizeVal)
            print('RAM size:  %d Kbytes' % ramSizeVal)
            print('Checksum complement:  0x%04x' % csumComp)
            print('Checksum           :  0x%04x' % csum)

        print('上記のヘッダ情報を参考にROM読み出しを行います。')
        
        # ファイル名入力
        fileNameStr = df_Common.inputFilename('作成ファイル名を入力してください。: ')

        fileAry = []
        addr = 0
        dumpSize1time = 1024
        if mapModeVal == 0 or mapModeVal == 2 or mapModeVal == 3:
            isLoROM = 1
        else:
            isLoROM = 0
        
        while addr < romSizeVal * 1024:
            # ファイル出力用配列に追加
            if isLoROM == 1:
                realAddr = df_SFC.calcRealAddrSFC_LoROM(addr)
            else:
                realAddr = df_SFC.calcRealAddrSFC_HiROM(addr)
            
            #print('\rアドレス {:06x} から読み出し中...'.format(realAddr), end='')
            print('\r[%.2f%%]アドレス {:06x} から読み出し中...'.format(realAddr) % ((addr / (romSizeVal * 1024))*100), end='')
            (retCode, retAry) = df_Common.readSizeByType(s, 0x07, realAddr, dumpSize1time)
            if retCode == 0:
                fileAry.extend(retAry)
                addr = addr + dumpSize1time
            else:
                print('リトライ実施(retCode: %d)' % retCode)
                #s.rtscts = 1
                continue

        print('')
        # チェックサム表示
        romSum = df_SFC.calcCheckSum_SFC(fileAry)
        print('読み出し結果から計算したチェックサム: 0x%04x' % romSum)
        
        
        # 配列をファイル出力
        with open(fileNameStr, "wb") as fOut:
            for x in fileAry:
                fOut.write(struct.pack("B", x))
        print('ファイル: %sを出力しました。' % fileNameStr)
        print('')

#------------------------------------------------
#
# メニューアイテム
# [SFM]Wakeup＆全体吸出し＆ROMファイル作成
# ※マップ情報を除く
#
#------------------------------------------------
def MenuItem_SFC_SFM_ReadROMAndCreateFile_ALL(s, connected):
    if connected != 1:
        print('カセットが未接続のため処理できません。')
    else:
        print('[SFM]Wakeup＆全体吸出し＆ROMファイル作成')
        
        # SFメモリwakeup処理
        df_SFM.wakeup_SFM(s)
        
        print('0.3秒待機します。')
        time.sleep(0.3)

        # Wakeup状態チェック
        (retCode, retAry) = df_Common.readSizeByType(s, 0x07, 0x2400, 16)
        if retCode != 0:
            print('Err: Wakeup状態読み取り処理失敗。')
            return g_menu
        else:
            if retAry[0] == 0x2a:
                print('SFメモリカセットのWakeup処理に成功しました。')
            else:
                print('Err: SFメモリカセットのWakeup処理に失敗しました。')
                return g_menu
                
        # ここまで来たら、Wakeup(HIMEM:ALL)に成功している
        # HIMEM×64バンク分を吸い出す

        # ファイル名入力
        fileNameStr = df_Common.inputFilename('作成ファイル名を入力してください。: ')

        fileAry = []
        addr = 0
        dumpSize1time = 1024
        allSize = 8 * 512 *1024    # SFメモリカセットは8 * 512Kbyte(4Mbit) = 4096Kbyte

        while addr < allSize:
            # ファイル出力用配列に追加
            realAddr = df_SFC.calcRealAddrSFC_HiROM(addr)
            
            #print('\rアドレス {:06x} から読み出し中...'.format(realAddr), end='')
            print('\r[%.2f%%]アドレス {:06x} から読み出し中...'.format(realAddr) % ((addr / allSize)*100), end='')

            (retCode, retAry) = df_Common.readSizeByType(s, 0x07, realAddr, dumpSize1time)
            if retCode == 0:
                fileAry.extend(retAry)
                addr = addr + dumpSize1time
            else:
                print('リトライ実施(retCode: %d)' % retCode)
                #s.rtscts = 1
                continue
        
        # 配列をファイル出力
        with open(fileNameStr, "wb") as fOut:
            for x in fileAry:
                fOut.write(struct.pack("B", x))
        print('ファイル: %sを出力しました。' % fileNameStr)
        print('')

#------------------------------------------------
#
# メニューアイテム
# [SFM]Wakeup＆ゲーム選択吸出し＆ROMファイル作成
#
#------------------------------------------------
def MenuItem_SFC_SFM_ReadROMAndCreateFile_Selected(s, connected):
    if connected != 1:
        print('カセットが未接続のため処理できません。')
    else:
        print('[SFM]Wakeup＆ゲーム選択吸出し＆ROMファイル作成')
        
        # SFメモリwakeup処理
        df_SFM.wakeup_SFM(s)
        
        print('0.3秒待機します。')
        time.sleep(0.3)

        # Wakeup状態チェック
        (retCode, retAry) = df_Common.readSizeByType(s, 0x07, 0x2400, 16)
        if retCode != 0:
            print('Err: Wakeup状態読み取り処理失敗。')
            return
        else:
            if retAry[0] == 0x2a:
                print('SFメモリカセットのWakeup処理に成功しました。')
            else:
                print('Err: SFメモリカセットのWakeup処理に失敗しました。')
                return
                
        # ここまで来たら、Wakeup(HIMEM:ALL)に成功している

        bankSizes = []

        # MENUプログラム存在チェック
        (retCode, retAry) = df_Common.readSizeByType(s, 0x07, 0xc07fc0, 12)
        #df_Common.dump(retAry)
        
        if retCode != 0:
            print('Err: 読み取り処理失敗。')
            return
        
        # 'MENU PROGRAM'か？
        if bytes(retAry) == bytes([0x4D, 0x45, 0x4E, 0x55, 
                                   0x20, 0x50, 0x52, 0x4F, 
                                   0x47, 0x52, 0x41, 0x4D]):
            print('MENU PROGRAMを検知しました。')
            hasMenu = 1
        else:
            print('MENU PROGRAMが存在しないか、検知できませんでした。')
            hasMenu = 0
    
        # もしメニューありなら、ゲーム一覧表示して、ゲームを選ばせる
        # メニュー無しなら、強制的に0
        if hasMenu == 1:
            print('MENU PROGRAMからディレクトリ情報を取得します。')
            
            # ディレクトリ一覧表示
            print('#  F  B Gamecode     タイトル                                     Date       Time     LAW/NIN ')
            print('- -- -- ------------ -------------------------------------------- ---------- -------- --------')
            #      0 00 00 123456789012 ０１２３４５６７８９０１２３４５６７８９０１ yyyy/mm/dd hh:mm:ss LAWxxxxx
            i = 0
            addr = 0xc60000
            while i < 8:
                # ディレクトリ位置0の情報を持ってくる
                (retCode, retAry) = df_Common.readSizeByType(s, 0x07, addr, 512)
                
                if retCode != 0:
                    print('Err: ディレクトリ情報参照失敗。')
                    return
                else:
                    if retAry[0] >= 8:
                        # ディレクトリ情報はもう無い
                        break

                    if retAry[0] != i:
                        print('Err: ディレクトリ情報が不正です。(メニュー情報#%d不正)' % i)
                        return
                    
                    # 各種情報の計算
                    dirIdx = retAry[0]
                    numOfFlashBlock = (retAry[4] * 0x100 + retAry[3]) / 4
                    numOfSRAMBlock  = (retAry[6] * 0x100 + retAry[5]) / 16
                    gamecode = str(bytes(retAry[7:19]), encoding='utf-8', errors='replace')
                    sjisName = str(bytes(retAry[0x13:0x3f]), encoding='sjis', errors='replace')
                    dateStr  = str(bytes(retAry[0x1bf:0x1c9]), encoding='utf-8', errors='replace')
                    timeStr  = str(bytes(retAry[0x1c9:0x1d1]), encoding='utf-8', errors='replace')
                    lawnin   = str(bytes(retAry[0x1d1:0x1d9]), encoding='utf-8', errors='replace')
                    
                    bankSizes.append(numOfFlashBlock)
                    
                    print('%1d %2d %2d %12s %s %s %s %s' % 
                        (dirIdx, numOfFlashBlock, numOfSRAMBlock, gamecode, sjisName,
                         dateStr, timeStr, lawnin))
                    
                    # 次のディレクトリ情報位置へ
                    i = i + 1
                    addr = addr + 0x2000
            # ディレクトリ表示 ここまで

            # ゲーム番号入力
            while True:
                print('ゲーム番号を入力してください。(#欄の数値、0-7): ', end='')
                gameNumStr = input()
                if len(gameNumStr) == 0:
                    continue
                
                gameNumInt = int(gameNumStr, 0)
                if gameNumInt < 0 or gameNumInt > 7:
                    continue
                break
            
            # 開始バンク番号計算
            startBankNum = 0
            i = 0
            while i < gameNumInt:
                startBankNum = startBankNum + int(bankSizes[i] * 8)
                i = i + 1
            
            # バンクサイズ
            banksToRead = int(bankSizes[gameNumInt] * 8)
            
        else:
            print('MENU PROGRAMからディレクトリ情報を取得できません。')
            gameNumInt   = 0
            startBankNum = 0
            banksToRead  = 64

        print('吸出し開始バンク番号:0x%02x' % startBankNum)
        print('バンク数            :0x%02x' % banksToRead)

        # HIMEMとして吸い出す

        # ファイル名入力
        while True:
            print('作成ファイル名を入力してください。: ', end='')
            fileNameStr = input()
            if len(fileNameStr) == 0:
                continue
            break

        fileAry = []
        addrStart = startBankNum * 0x10000
        addr = addrStart
        dumpSize1time = 1024
        maxAddr = (startBankNum * 0x10000) + (banksToRead * 0x10000)

        while addr < maxAddr:
            # ファイル出力用配列に追加
            realAddr = df_SFC.calcRealAddrSFC_HiROM(addr)
            
            #print('\rアドレス {:06x} から読み出し中...'.format(realAddr), end='')
            print('\r[%.2f%%]アドレス {:06x} から読み出し中...'.format(realAddr) % (((addr - addrStart) / (maxAddr - addrStart))*100), end='')
            (retCode, retAry) = df_Common.readSizeByType(s, 0x07, realAddr, dumpSize1time)
            if retCode == 0:
                fileAry.extend(retAry)
                addr = addr + dumpSize1time
            else:
                print('リトライ実施(retCode: %d)' % retCode)
                #s.rtscts = 1
                continue
        
        # 配列をファイル出力
        with open(fileNameStr, "wb") as fOut:
            for x in fileAry:
                fOut.write(struct.pack("B", x))
        print('ファイル: %sを出力しました。' % fileNameStr)
        print('')

#------------------------------------------------
#
# メニューアイテム
# [SFM]Wakeup＆マップ情報ファイル保存
#
#------------------------------------------------
def MenuItem_SFC_SFM_ReadMapInfoAndSaveFile(s, connected):
    if connected != 1:
        print('カセットが未接続のため処理できません。')
        return

    print('[SFM]Wakeup＆マップ情報ファイル保存')
    
    # SFメモリwakeup処理
    df_SFM.wakeup_SFM(s)
    
    print('0.3秒待機します。')
    time.sleep(0.3)

    # Wakeup状態チェック
    (retCode, retAry) = df_Common.readSizeByType(s, 0x07, 0x2400, 16)
    if retCode != 0:
        print('Err: Wakeup状態読み取り処理失敗。')
        return
    else:
        if retAry[0] == 0x2a:
            print('SFメモリカセットのWakeup処理に成功しました。')
        else:
            print('Err: SFメモリカセットのWakeup処理に失敗しました。')
            return
            
    # ここまで来たら、Wakeup(HIMEM:ALL)に成功している

    # Maker ID/Device IDの確認
    (mID_c0, dID_c0) = df_SFM.idFlash_SFM(s, 0xc0)
    if (mID_c0 != 0xc2) or (dID_c0 != 0xf3):
        print('Err: Flash %02x Maker ID/Device IDが不正です。(%02x%02x)' % (0xc0, mID_c0, dID_c0))
        return
    print('Flash %02x Maker ID/Device ID: %02x/%02x...OK' % (0xc0, mID_c0, dID_c0))

    (mID_e0, dID_e0) = df_SFM.idFlash_SFM(s, 0xe0)
    if (mID_e0 != 0xc2) or (dID_e0 != 0xf3):
        print('Err: Flash %02x Maker ID/Device IDが不正です。(%02x%02x)' % (0xe0, mID_e0, dID_e0))
        return
    print('Flash %02x Maker ID/Device ID: %02x/%02x...OK' % (0xe0, mID_e0, dID_e0))

    # リセット
    print('Flashをリセットし、0.1秒待機します。')
    df_SFM.resetFlash_SFM(s, 0xc0)
    df_SFM.resetFlash_SFM(s, 0xe0)
    time.sleep(0.1)
    
    # マップ情報のダンプ
    print('マップ情報（ダンプ）')
    df_SFM.printMapping_SFM(s)

    # マップ情報を配列に読み込む
    print('SFメモリカセットのマップ情報を読み込みます...', end = '')
    (retCode, fileAry) = df_SFM.readMapping_SFM(s)
    if retCode != 0:
        return
    print('完了。')

    # ファイル名入力
    fileNameStr = df_Common.inputFilename('作成ファイル名を入力してください。(推奨拡張子:*.map): ')
    
    # 配列をファイル出力
    with open(fileNameStr, "wb") as fOut:
        for x in fileAry:
            fOut.write(struct.pack("B", x))
    print('ファイル: %sを出力しました。' % fileNameStr)
    print('')


#------------------------------------------------
#
# メニューアイテム
# [SFM]全セクターのプロテクト状態表示
#
#------------------------------------------------
def MenuItem_SFC_SFM_DispSectorProtect(s, connected):
    if connected != 1:
        print('カセットが未接続のため処理できません。')
    else:
        print('[SFM]全セクターのプロテクト状態表示')
    
    # SFメモリwakeup処理
    df_SFM.wakeup_SFM(s)
    
    print('0.3秒待機します。')
    time.sleep(0.3)

    # Wakeup状態チェック
    (retCode, retAry) = df_Common.readSizeByType(s, 0x07, 0x2400, 16)
    if retCode != 0:
        print('Err: Wakeup状態読み取り処理失敗。')
        return
    else:
        if retAry[0] == 0x2a:
            print('SFメモリカセットのWakeup処理に成功しました。')
        else:
            print('Err: SFメモリカセットのWakeup処理に失敗しました。')
            return
    
    df_SFM.readSectorProtection_SFM(s, 0xc0)
    df_SFM.readSectorProtection_SFM(s, 0xe0)
    
#------------------------------------------------
#
# メニューアイテム
# [SFM] マップ情報イレーズコマンド送信
#
#------------------------------------------------
def MenuItem_SFC_SFM_EraseMapInfo(s, connected):
    if connected != 1:
        print('カセットが未接続のため処理できません。')
    else:
        print('[SFM] マップ情報イレーズコマンド送信')
    
    df_SFM.resetFlash_SFM(s, 0xc5)
    df_SFM.resetFlash_SFM(s, 0xe0)
    
    df_SFM.sendHiddenSectorEraseCommand_SFM(s, 0xc2)
    df_SFM.sendHiddenSectorEraseCommand_SFM(s, 0xe0)

    print('Enterキーを押してください。')
    input()

#------------------------------------------------
#
# メニューアイテム
# [SFM] c0/e0リセットコマンド送信
#
#------------------------------------------------
def MenuItem_SFC_SFM_Reset(s, connected):
    if connected != 1:
        print('カセットが未接続のため処理できません。')
    else:
        print('[SFM] c0/e0リセットコマンド送信')
    
    df_SFM.resetFlash_SFM(s, 0xc0)
    df_SFM.resetFlash_SFM(s, 0xe0)

#------------------------------------------------
#
# メニューアイテム
# [SFM]Flash c0セクタープロテクト有効化
#
#------------------------------------------------
def MenuItem_SFC_SFM_SectorProtectC0(s, connected):
    if connected != 1:
        print('カセットが未接続のため処理できません。')
    else:
        print('[SFM]Flash c0セクタープロテクト有効化')
    
    i = 0
    while i < 0x1f:
        df_SFM.setSectorProtect_SFM(s, 0xc0 + i)
        i = i + 1

#------------------------------------------------
#
# メニューアイテム
# [SFM]Flash c0プロテクト有効化
#
#------------------------------------------------
def MenuItem_SFC_SFM_SetFlashProtect(s, connected):
    if connected != 1:
        print('カセットが未接続のため処理できません。')
    else:
        print('[SFM]Flash c0プロテクト有効化')
    
    df_SFM.setFlashProtect_SFM(s, 0xc0)
    
    print('Enterキーを押してください。')
    input()
    
    
#------------------------------------------------
#
# メニューアイテム
# [SFM]Flash c0プロテクト無効化
#
#------------------------------------------------
def MenuItem_SFC_SFM_UnsetFlashProtect(s, connected):
    if connected != 1:
        print('カセットが未接続のため処理できません。')
    else:
        print('[SFM]Flash c0プロテクト無効化')
    
    df_SFM.unsetFlashProtect_SFM(s, 0xc0)
    
    print('Enterキーを押してください。')
    input()

#------------------------------------------------
#
# メニューアイテム
# [SFM]Flash c0末尾セクタープロテクト無効化
#
#------------------------------------------------
def MenuItem_SFC_SFM_SectorUnprotectC0End(s, connected):
    if connected != 1:
        print('カセットが未接続のため処理できません。')
    else:
        print('[SFM]Flash c0セクタープロテクト有効化')
    
    i = 0x1e
    while i < 0x20:
        df_SFM.unsetSectorProtect_SFM(s, 0xc0 + i)
        i = i + 1


#------------------------------------------------
#
# メニューアイテム
# [SFM]マップ情報書き込み
#
#------------------------------------------------
def MenuItem_SFC_SFM_WriteMAPInfo(s, connected):
    if connected != 1:
        print('カセットが未接続のため処理できません。')
        return

    print('[SFM]マップ情報書き込み')
    
    # SFメモリwakeup処理
    df_SFM.wakeup_SFM(s)
    
    print('0.3秒待機します。')
    time.sleep(0.3)

    # Wakeup状態チェック
    if df_SFM.isWakeup_SFM(s) != True:
        print('Err: SFメモリカセットがWakeup状態ではありません。')
        return

    print('※※※※※※※※※※※※※※※※※※※※※※※※※※※※※')
    print('警告：接続中のSFメモリカセットのマップ情報を書き換えます。')
    print('      失敗するとカセットは正常に動作しなくなります。')
    print('※※※※※※※※※※※※※※※※※※※※※※※※※※※※※')
    while True:
        print('SFメモリカセットの内容を書き換えますか？(y/n):', end='')
        inputStr = input()
        if   inputStr == 'y' or inputStr == 'Y':
            break
        elif inputStr == 'n' or inputStr == 'N':
            return
    
    # なぜマップ情報の操作でeraseMapping(s, 0xc0)でなく、
    # eraseMapping(s, 0xc2)を使うのか・・・
    #   セクタープロテクト状況を見ると、出荷時（Loppi書き換え後？）は
    #   0xc0&0xc1にはセクタープロテクトがかけられた状態であるため。
    #   なんとなく0xc2を使った方が良いと思った。
    #   0xc0でも0xc2でも、実際には同じFlashに対する操作になるので影響はない。

    # マップ情報の消去
    print('マップ情報を消去します...', flush=True, end='')
    #eraseMapping(s, 0xd0)       # NP.inoではこうなっている
    if df_SFM.eraseMapping(s, 0xc2) != True:
        print('Err: Flash c0のマップ情報消去に失敗。')
        return
        
    if df_SFM.eraseMapping(s, 0xe0) != True:
        print('Err: Flash e0のマップ情報消去に失敗。')
        return

    print('完了。')
    

    # マップ情報のブランクチェック
    print('マップ情報領域のブランクチェック...', flush=True)
    if df_SFM.blankcheckMapping_SFM(s) == 0:
        print('NG。ブランクチェックに失敗しました。')
        input()
        return
    print('OK。')

    #       ファイル名入力
    mapFileNameStr = df_Common.inputFilename(
        '書き込みを行うマップ情報ファイル名を入力してください：')
    
    #       ファイル読み込み
    with open(mapFileNameStr, "rb") as fIn:
        readStr = fIn.read()
        readBytes = bytes(readStr)
        if len(readBytes) > 0:
            print('ファイルを読み込みました。')
    
    #       マップ情報書き換え
    df_SFM.writeMapping_SFM(s,0xc2, readBytes, 0);
    df_SFM.writeMapping_SFM(s,0xe0, readBytes, 256);

    #       マップ情報ベリファイ
    print('マップ情報の書き換え完了。ベリファイします。')
    (retCode, currentBytes) = df_SFM.readMapping_SFM(s)
    if retCode != 0:
        print('NG。カセットからの読み込みに失敗しました。')
        input()
        return
    #if currentBytes != readBytes:  # <--- ダメ
    if df_Common.verifyBytes(currentBytes,readBytes) == False:
        print('NG。ベリファイに失敗しました。')
        input()
        return
    print('OK。ベリファイに成功しました。')

#------------------------------------------------
#
# メニューアイテム
# [SFM]SFメモリカセット用ROMイメージ書き込み
#
#------------------------------------------------
def MenuItem_SFC_SFM_WriteROMImage(s, connected):
    if connected != 1:
        print('カセットが未接続のため処理できません。')
        return g_menu

    print('[SFM]Wakeup＆SFメモリカセット用イメージ書き込み')
    
    # SFメモリwakeup処理
    df_SFM.wakeup_SFM(s)
    
    print('0.3秒待機します。')
    time.sleep(0.3)

    # Wakeup状態チェック
    if df_SFM.isWakeup_SFM(s) != True:
        print('Err: SFメモリカセットがWakeup状態ではありません。')
        return

    print('※※※※※※※※※※※※※※※※※※※※※※※※※※※※※')
    print('警告：接続中のSFメモリカセットを書き換えます。')
    print('      失敗するとカセットは正常に動作しなくなります。')
    print('※※※※※※※※※※※※※※※※※※※※※※※※※※※※※')
    while True:
        print('SFメモリカセットの内容を書き換えますか？(y/n):', end='')
        inputStr = input()
        if   inputStr == 'y' or inputStr == 'Y':
            break
        elif inputStr == 'n' or inputStr == 'N':
            return

    # ここまで来たら、Wakeup(HIMEM:ALL)に成功している

    # ライトプロテクト解除
    print('SFメモリカセットのライトプロテクトを解除します。')
    if df_SFM.unprotectCart_SFM(s) == False:
        print('失敗。')
        input()
        return
    print('成功。')

    # ファイル読み込み
    #       ファイル名入力
    fileNameStr = df_Common.inputFilename('書き込みを行うファイル名を入力してください：')
    
    #       ファイル読み込み
    with open(fileNameStr, "rb") as fIn:
        readStr = fIn.read()
        readBytes = bytes(readStr)
        if len(readBytes) > 0:
            print('ファイルを読み込みました。')
    
    # 書き換え
    #       Flash c0
    bankNum = 0xc0
    print('Flash %02xの書き換えを実行します。' % bankNum)
    if df_SFM.write_SFM(s, bankNum, readBytes, 0) == False:
        print('Err: Flash %02xの書き換えに失敗しました。' % bankNum)
        input()
        return
    print('Flash %02xの書き換えに成功しました。' % bankNum)
    
    #       ベリファイ
    print('Flash %02xのベリファイを実行します。' % bankNum)
    (retCode, currentBytes) = df_SFM.readFlash_SFM(s, 0x000000)
    if retCode == False:
        print('Err: Flash %02xの読み込みに失敗しました。' % bankNum)
        input()
        return
    if df_Common.verifyBytes(
        currentBytes, readBytes[0:0x200000]) == False:
        print('Err: Flash %02xのベリファイに失敗しました。' % bankNum)
        input()
        return
    print('Flash %02xのベリファイに成功しました。' % bankNum)
        
        
    #       Flash e0
    bankNum = 0xe0
    
    print('Flash %02xの書き換えを実行します。' % bankNum)
    if df_SFM.write_SFM(s, bankNum, readBytes, 0x200000) == False:
        print('Err: Flash %02xの書き換えに失敗しました。' % bankNum)
        input()
        return
    print('Flash %02xの書き換えに成功しました。' % bankNum)
    
    #       ベリファイ
    print('Flash %02xのベリファイを実行します。' % bankNum)
    (retCode, currentBytes) = df_SFM.readFlash_SFM(s, 0x200000)
    if retCode == False:
        print('Err: Flash %02xの読み込みに失敗しました。' % bankNum)
        input()
        return
    if df_Common.verifyBytes(
        currentBytes, readBytes[0x200000:]) == False:
        print('Err: Flash %02xのベリファイに失敗しました。' % bankNum)
        input()
        return
    print('Flash %02xのベリファイに成功しました。' % bankNum)

